@@ -17,3 +17,14 @@
 void tomlxs_free_string(char *ptr);
 
 void tomlxs_free_timestamp(toml_timestamp_t *ptr);
+
+enum toml_xs_type {
+    TOML_XS_TYPE_INVALID,
+    TOML_XS_TYPE_TABLE,
+    TOML_XS_TYPE_ARRAY,
+    TOML_XS_TYPE_STRING,
+    TOML_XS_TYPE_BOOLEAN,
+    TOML_XS_TYPE_INTEGER,
+    TOML_XS_TYPE_DOUBLE,
+    TOML_XS_TYPE_TIMESTAMP,
+};
